/*
 * Database.java   E.L. 2006-10-19
 *
 * Dette er en "database-wrapper" for vedlikehold av tabellen Person.
 * Konstruktren pner en databaseforbindelse som holdes inntil klienten
 * kaller metoden kobleNedforbindelse().
 *
 * Feilhndtering: Konstruktren kaster unntaksobjekt dersom JDBC-klassen ikke kan
 * lastes, eller databaseforbindelsen ikke kan pnes. For vrig hndteres SQLException
 * ved en utskrift til konsollet. Slike unntak skyldes programmeringsfeil og sendes ikke
 * til klienten.
 */

import java.util.*;
import java.sql.*;
import mittBibliotek.Person;

public class Database {
  private Connection forbindelse;
  private int kode; // Returkode fra databasesystemet dersom primrnkkelkonflikt

  /*
   * Laster JDBC-klassene og pner databaseforbindelsen.
   */
  public Database(String dbDriver, String dbNavn, String brukernavn,
                  String passord, int feilkode) throws Exception {
    try {
      kode = feilkode;
      Class.forName(dbDriver);
      forbindelse = DriverManager.getConnection(dbNavn, brukernavn, passord);
    } catch (Exception e) {
      skrivMelding(e, "konstruktr");
      throw e;
    }
  }

  /*
   * Metoder for  lukke forbindelse. M kalles av klienten fr programmet avsluttes.
   */
  public void kobleNedForbindelse() {
    try {
      if (forbindelse != null) forbindelse.close();
      System.out.println("Har koblet ned en databaseforbindelse");
    } catch (SQLException e) {
      skrivMelding(e, "kobleNedForbindelse()");
    }
  }

  /*
   * Returnerer en tabell-liste som inneholder alle dataene i tabellen person.
   * Tabell-listen inneholder person-objekter. Listen er sortert etter etternavn.
   * Returnerer null dersom SQLException er kastet.
   */
  public ArrayList<Person> finnAlle() {
    ArrayList<Person>  alle = new ArrayList<Person>();
    /* initcap() er en sql-funksjon som setter frste bokstav i hvert ord stor, resten sm */
    String sqlsetning = "select persnr, initcap(fornavn) as fn, initcap(etternavn) as en " +
                        "from person order by etternavn, fornavn";
    System.out.println(sqlsetning);

    ResultSet res = null;
    Statement setning = null;
    try {
      setning = forbindelse.createStatement();
      res = setning.executeQuery(sqlsetning);
      while (res.next()) {
        int nr = res.getInt("persnr");
        String fornavn = res.getString("fn");
        String etternavn = res.getString("en");
        Person navnet = new Person(nr, fornavn, etternavn);
        alle.add(navnet);
      }
    } catch (SQLException e) {
      skrivMelding(e, "finnAlle()");
      alle = null;
    } finally {  // setningene nedenfor utfres alltid
      lukkResSet(res);
      lukkSetning(setning);
    }
    return alle;
  }

  /*
   * Endrer navn. Det nye navnet samt personens nummer sendes inn som argument.
   * Returnerer false dersom ingen person med dette nummeret eksisterer,
   * eller SQLException er inntruffet.
   */
  public boolean endreNavn(Person personen) {
    int nr = personen.finnPersonNr();
    String nyttFornavn = personen.finnFornavn();
    String nyttEtternavn = personen.finnEtternavn();
    String sqlsetning = "update person set fornavn = upper('" + nyttFornavn
          + "'), etternavn = upper('" + nyttEtternavn + "') where persnr = " + nr;
    System.out.println(sqlsetning);

    Statement setning = null;
    try {
      setning = forbindelse.createStatement();
      if (setning.executeUpdate(sqlsetning) == 0) return false;
      else return true;
    } catch (SQLException e) {
      skrivMelding(e, "endreNavn()");
    } finally {
      lukkSetning(setning);  // utfres selv om det er return-setninger rett over
    }
    return false;
  }

  /*
   * Registrerer ny person.
   * Nummer tildeles av systemet, og et fullstendig personobjekt returneres til klienten.
   * Dersom tabellen er tom, settes personnr lik 1, ellers settes det lik
   * hittil strste nummer + 1.(P grunn av at data kan slettes,
   * kan denne metoden fre til hull i nummersekvensen.)
   * Metoden returnerer null dersom feil oppsttt.
   */
  public Person registrerNyPerson(String fornavn, String etternavn) {
    int nyttNr = 1; // dersom det ikke er data i databasen
    boolean ok = true;
    do {
      ResultSet res = null;
      Statement setning = null;
      try {
        String sqlsetning = "select max(persnr) as maks from person";
        setning = forbindelse.createStatement();
        res = setning.executeQuery(sqlsetning);
         /*Hvis det ikke er data i tabellen, vil maks() returnere SQL NULL. getInt() vil omforme denne til 0,
         * og nyttNr blir dermed 1 dersom det ikke er data i databasen. Det er ok.  */
        res.next();
        nyttNr = res.getInt("maks") + 1;
        sqlsetning = "insert into person values(" + nyttNr + ", upper('"
              + fornavn + "'), upper('" + etternavn + "'))";
        System.out.println(sqlsetning);
        setning.executeUpdate(sqlsetning);
        ok = true;
      } catch (SQLException e) {
        /* Dersom feilen skyldes at person med samme nummer eksisterer fra fr,
           m det bety at en annen klient har vrt inne i databasen mellom
           vre to sql-setninger. Vi kjrer derfor gjennom en gang til.
           (Dette er noe som vil inntreffe svrt sjelden.) */
        if (e.getErrorCode() == kode) ok = false;
        else {
          skrivMelding(e, "registrerNyPerson()");
          return null;   // RETUR, feil oppsttt (finally-blokken blir utfrt)
        }
      } finally {
        lukkResSet(res);
        lukkSetning(setning);
      }
    } while (!ok);
    return new Person(nyttNr, fornavn, etternavn);
  }

  /*
   * Sletter person med gitt nummer.
   * Returnerer false dersom personen ikke finnes, eller SQLException inntruffet.
   */
  public boolean slettPerson(int nr) {
    String sqlsetning = "delete from person where persnr = " + nr;
    System.out.println(sqlsetning);

    Statement setning = null;
    try {
      setning = forbindelse.createStatement();
      if (setning.executeUpdate(sqlsetning) == 0) return false;
      else return true;
    } catch (SQLException e) {
      skrivMelding(e, "slettPerson()");
    } finally {
      lukkSetning(setning);
    }
    return false;
  }

  /*
   * Private metoder for  lukke Statement- og ResultSet-objekt
   */
  private void lukkResSet(ResultSet res) {
    try {
      if (res != null) res.close();
    } catch (SQLException e) {
      skrivMelding(e, "lukkResSet()");
    }
  }

  private void lukkSetning(Statement stm) {
    try {
      if (stm != null) stm.close();
    } catch (SQLException e) {
      skrivMelding(e, "lukkSetning()");
    }
  }

  private void skrivMelding(Exception e, String metode) {
    System.err.println("*** Feil oppsttt i klassen Database, metode " + metode + ". ***");
    e.printStackTrace(System.err);
  }
}